"""Song status change

Song change monitoring and XML reporting class
"""

import os
import time
import base64
import appscript

from lxml.builder import E

from . import MusicPlayerError
from .client import Client
from systematic.shell import normalized

INFO_KEY_ORDER = (
    'path',
    'album_artist',
    'artist',
    'album',
    'name',
    'comment',
    'length',
    'genre',
    'year',
    'track_number',
    'track_count'
)

IGNORE_TRACK_FIELDS = (
    'persistent_ID',
    'id',
    'index',
    'start',
    'finish',
    'database_ID',
    'time',
    'album_rating_kind',
    'compilation',
    'date_added',
    'enabled',
    'gapless',
    'modification_date',
    'played_count',
    'podcast',
    'rating_kind',
    'release_date',
    'played_date',
    'skipped_date',
    'skipped_count',
    'bookmarkable',
    'shufflable',
    'unplayed',
    'video_kind',
    'location',
)


class musicPlayerStatus(object):
    """Song status change monitoring

    This class can monitor status of currently playing track and
    return track change information with next() iterator, when current
    track changes.
    """

    def __init__(self, interval=1, xml_output=False, export_albumart=False):
        self.interval = int(interval)
        self.xml_output = xml_output
        self.export_albumart = export_albumart

        self.__last_checked__ = None

        try:
            self.client = Client()
            try:
                self.current = self.client.current_track
                self.status = self.client.status
            except appscript.reference.CommandError:
                self.current = None
        except MusicPlayerError:
            self.client = None
            self.current = None
            self.status = None

    def __iter__(self):
        return self

    def next(self):
        """Return song info on song change

        Iterate current object, sleeping self.interval time and returning
        status and XML formatted track information when song is changed.
        """
        while True:
            try:
                if self.client is None:
                    self.client = Client()

                n = self.client.current_track
                if self.client.status != self.status:
                    self.status = self.client.status

                    if self.status == 'playing':
                        return self.status, self.songinfo(n)
                    else:
                        return self.status, None

                if n is not None:
                    if self.current is None or n.id != self.current.id:
                        self.current = n
                        return self.client.status, self.songinfo(n)

            except MusicPlayerError:
                self.client = None
                pass
            except appscript.reference.CommandError:
                pass

            time.sleep(self.interval)

    def songinfo(self, track=None, xml_output=False, export_albumart=False):
        """Current playing track info

        Return current playing track information as XML or None if no
        track was selected.
        """

        try:
            if self.client.status in ('stopped', 'paused'):
                return None

            if not track:
                track = self.client.current_track

            started = time.mktime(time.localtime()) - self.client.player_position()
            info = {
                'path': track.path,
                'started': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(started)),
                'album_artist':  track.album_artist,
                'artist':  track.artist,
                'album': track.album,
                'name': track.name,
                'comment': track.comment,
                'genre': track.genre,
                'length': track.time,
                'track_number': track.track_number,
                'track_count': track.track_count,
            }

            if track.year != 0:
                info['year'] = track.year

            albumart = os.path.join(os.path.dirname(normalized(track.path)), 'artwork.jpg')
            if os.path.isfile(albumart):
                info['artwork'] = os.path.realpath(albumart)

            if not self.xml_output and not xml_output:
                return info

            xml = E(
                'pytunes',
                persistent_ID=track.persistent_ID,
                started=info['started']
            )

            for k in INFO_KEY_ORDER:
                # Skip node attributes
                if k in ['started']:
                    continue

                # Optional fields
                if k not in info:
                    continue

                # Don't export empty fields
                if info[k] == '':
                    continue

                v = info[k]
                try:
                    if isinstance(v, int):
                        if v == 0:
                            continue
                        xml.append(E(k, '{:d}'.format(v)))
                    elif isinstance(v, float):
                        if v == 0:
                            continue
                        xml.append(E(k, '{:3.2f}'.format(v)))
                    elif v != '':
                        xml.append(E(k, v))

                except ValueError:
                    print('ERROR encoding key {} type {}: {}'.format(k, type(v), v))

            for k in sorted(track.keys()):
                v = track[k]
                if k in INFO_KEY_ORDER or k in IGNORE_TRACK_FIELDS:
                    continue

                try:
                    if isinstance(v, int):
                        if v == 0:
                            continue
                        xml.append(E(k, '{0:d}'.format(v)))
                    elif isinstance(v, float):
                        if v == 0:
                            continue
                        xml.append(E(k, '{0:3.2f}'.format(v)))
                    else:
                        if v == '':
                            continue
                        xml.append(E(k, v))

                except TypeError:
                    print('ERROR encoding key {} type {}: {}'.format(k, type(v), v))

                except ValueError:
                    print('ERROR encoding key {} type {}: {}'.format(k, type(v), v))

            if (self.export_albumart or export_albumart) and albumart:
                xml.append(E(
                    'albumart',
                    base64.b64encode(open(albumart, 'r').read()),
                    filename=os.path.basename(albumart)
                ))

            return xml

        except MusicPlayerError:
            self.client = None
            return None




import os
import appscript
import mactypes

from . import MusicPlayerError


class Playlist(object):
    """
    Music player playlist

    Abstraction of a single MacOS music player playlist
    """

    def __init__(self, client, name=None):
        self.client = client
        self.__index__ = 0

        try:
            if name is None:
                self.playlist = self.client.get(self.client.library_playlists['library'])
                name = 'library'
            else:
                self.playlist = self.client.get(self.client.user_playlists[name])
        except appscript.reference.CommandError:
            raise MusicPlayerError('No such playlist: {}'.format(name))

        self.name = name
        self.__update_len__()

    def __update_len__(self):
        try:
            self.__len_cached__ = self.client.get(self.playlist.file_tracks[-1].index)
        except appscript.reference.CommandError:
            self.__len_cached__ = 0

    def __repr__(self):
        return 'playlist:{}'.format(self.name)

    def __len__(self):
        return self.__len_cached__

    def __str__(self):
        return self.name

    @property
    def parent(self):
        """
        Parent playlist

        Return parent playlist or playlist folder
        """
        try:
            return self.playlist.parent.get().name.get()
        except appscript.reference.CommandError:
            return None

    @property
    def is_smart(self):
        """Is this smart playlist

        Return true if this is a smart playlist

        """
        return self.playlist.smart.get() and True or False

    @property
    def path(self):
        """Playlist path

        Return relative path for this playlist

        """
        path = [self.name]
        try:
            p = self.playlist.parent.get()
            while True:
                path.append(p.name.get())
                p = p.parent.get()

        except appscript.reference.CommandError:
            pass

        path.reverse()
        return os.sep.join(path)

    def __getattr__(self, attr):
        try:
            ref = getattr(self.playlist, attr)
            return self.client.get(ref)

        except AttributeError:
            pass

        except appscript.reference.CommandError:
            pass

        raise AttributeError('No such playlist item: {}'.format(attr))

    def __getitem__(self, index):
        """Get track by index

        """
        try:
            index = int(index)
        except ValueError:
            raise ValueError('Invalid playlist index: {}'.format(index))

        if index < 0:
            index = len(self) - index

        try:
            index = int(index) + 1
            if index > len(self):
                raise StopIteration
            return self.client.get_playlist_track(self.playlist, index)
        except ValueError:
            self.__update_len__()
            raise ValueError('Invalid playlist index: {}'.format(index))
        except appscript.reference.CommandError:
            self.__update_len__()
            raise IndexError('Out of playlist: {:d}'.format(index))

    def __iter__(self):
        return self

    def __next__(self):
        return self.next() # noqa B305

    def next(self):
        """Iterate playlist

        Iterate tracks on playlist
        """
        try:
            if self.__index__ > len(self):
                raise StopIteration
            try:
                track = self[self.__index__]
                self.__index__ += 1
            except IndexError:
                raise StopIteration
        except StopIteration:
            self.__index__ = 0
            raise StopIteration

        return track

    def jump(self, index):
        """Jump to position

        Jump to playlist position

        """
        if index < 0 or index > len(self):
            raise ValueError
        self.__index__ = index + 1
        return self[self.__index__]

    def add(self, files):
        """Add files to playlist

        Add provided files to the playlist

        """
        if not isinstance(files, list):
            files = [files]
        self.client.add([mactypes.Alias(entry) for entry in files], to=self.playlist)
        self.__update_len__()

    def delete(self, entry):
        """Delete entry from playlist

        Delete provided track entry from playlist (via entry.track)

        """
        try:
            self.client.delete(entry.track)
            if self.__index__ > 0:
                self.__index__ -= 1
            self.__update_len__()
        except appscript.reference.CommandError as e:
            raise MusicPlayerError('Error deleting track {}: {}'.format(entry.track, e))



"""
iTunes / Music appscript terminology from osascript
"""

version = 1.1
path = '/Applications/iTunes.app'

classes = [
    ('print_settings', b'pset'),
    ('AirPlay_device', b'cAPD'),
    ('application', b'capp'),
    ('double_integer', b'comp'),
    ('picture', b'PICT'),
    ('artwork', b'cArt'),
    ('audio_CD_playlist', b'cCDP'),
    ('audio_CD_track', b'cCDT'),
    ('browser_window', b'cBrW'),
    ('encoder', b'cEnc'),
    ('EQ_preset', b'cEQP'),
    ('EQ_window', b'cEQW'),
    ('file_track', b'cFlT'),
    ('folder_playlist', b'cFoP'),
    ('item', b'cobj'),
    ('library_playlist', b'cLiP'),
    ('miniplayer_window', b'cMPW'),
    ('playlist', b'cPly'),
    ('playlist_window', b'cPlW'),
    ('radio_tuner_playlist', b'cRTP'),
    ('shared_track', b'cShT'),
    ('source', b'cSrc'),
    ('subscription_playlist', b'cSuP'),
    ('track', b'cTrk'),
    ('URL_track', b'cURT'),
    ('user_playlist', b'cUsP'),
    ('video_window', b'cNPW'),
    ('visual', b'cVis'),
    ('window', b'cwin')
]

enums = [
    ('track_listing', b'kTrk'),
    ('album_listing', b'kAlb'),
    ('cd_insert', b'kCDi'),
    ('standard', b'lwst'),
    ('detailed', b'lwdt'),
    ('stopped', b'kPSS'),
    ('playing', b'kPSP'),
    ('paused', b'kPSp'),
    ('fast_forwarding', b'kPSF'),
    ('rewinding', b'kPSR'),
    ('off', b'kRpO'),
    ('one', b'kRp1'),
    ('all', b'kAll'),
    ('songs', b'kShS'),
    ('albums', b'kShA'),
    ('groupings', b'kShG'),
    ('small', b'kVSS'),
    ('medium', b'kVSM'),
    ('large', b'kVSL'),
    ('library', b'kLib'),
    ('iPod', b'kPod'),
    ('audio_CD', b'kACD'),
    ('MP3_CD', b'kMCD'),
    ('radio_tuner', b'kTun'),
    ('shared_library', b'kShd'),
    ('iTunes_Store', b'kITS'),
    ('unknown', b'kUnk'),
    ('albums', b'kSrL'), ('artists', b'kSrR'),
    ('composers', b'kSrC'),
    ('displayed', b'kSrV'),
    ('songs', b'kSrS'),
    ('none', b'kNon'),
    ('Books', b'kSpA'),
    ('folder', b'kSpF'),
    ('Genius', b'kSpG'),
    ('iTunes_U', b'kSpU'),
    ('Library', b'kSpL'),
    ('Movies', b'kSpI'),
    ('Music', b'kSpZ'),
    ('Podcasts', b'kSpP'),
    ('Purchased_Music', b'kSpM'),
    ('TV_Shows', b'kSpT'),
    ('alert_tone', b'kMdL'),
    ('audiobook', b'kMdA'),
    ('book', b'kMdB'),
    ('home_video', b'kVdH'),
    ('iTunesU', b'kMdI'),
    ('movie', b'kVdM'),
    ('song', b'kMdS'),
    ('music_video', b'kVdV'),
    ('podcast', b'kMdP'),
    ('ringtone', b'kMdR'),
    ('TV_show', b'kVdT'),
    ('voice_memo', b'kMdO'),
    ('user', b'kRtU'),
    ('computed', b'kRtC'),
    ('computer', b'kAPC'),
    ('AirPort_Express', b'kAPX'),
    ('Apple_TV', b'kAPT'),
    ('AirPlay_device', b'kAPO'),
    ('unknown', b'kAPU'),
    ('purchased', b'kPur'),
    ('matched', b'kMat'),
    ('uploaded', b'kUpl'),
    ('ineligible', b'kRej'),
    ('removed', b'kRem'),
    ('error', b'kErr'),
    ('duplicate', b'kDup'),
    ('subscription', b'kSub'),
    ('no_longer_available', b'kRev'),
    ('not_uploaded', b'kUpP')
]

properties = [
    ('active', b'pAct'),
    ('address', b'pURL'),
    ('AirPlay_enabled', b'pAPE'),
    ('album', b'pAlb'),
    ('album_artist', b'pAlA'),
    ('album_disliked', b'pAHt'),
    ('album_loved', b'pALv'),
    ('album_rating', b'pAlR'),
    ('album_rating_kind', b'pARk'),
    ('artist', b'pArt'),
    ('available', b'pAva'),
    ('band_1', b'pEQ1'),
    ('band_10', b'pEQ0'),
    ('band_2', b'pEQ2'),
    ('band_3', b'pEQ3'),
    ('band_4', b'pEQ4'),
    ('band_5', b'pEQ5'),
    ('band_6', b'pEQ6'),
    ('band_7', b'pEQ7'),
    ('band_8', b'pEQ8'),
    ('band_9', b'pEQ9'),
    ('bit_rate', b'pBRt'),
    ('bookmark', b'pBkt'),
    ('bookmarkable', b'pBkm'),
    ('bounds', b'pbnd'),
    ('bpm', b'pBPM'),
    ('capacity', b'capa'),
    ('category', b'pCat'),
    ('class_', b'pcls'),
    ('closeable', b'hclb'),
    ('cloud_status', b'pClS'),
    ('collapseable', b'pWSh'),
    ('collapsed', b'wshd'),
    ('collating', b'lwcl'),
    ('comment', b'pCmt'),
    ('compilation', b'pAnt'),
    ('composer', b'pCmp'),
    ('container', b'ctnr'),
    ('converting', b'pCnv'),
    ('copies', b'lwcp'),
    ('current_AirPlay_devices', b'pAPD'),
    ('current_encoder', b'pEnc'),
    ('current_EQ_preset', b'pEQP'),
    ('current_playlist', b'pPla'),
    ('current_stream_title', b'pStT'),
    ('current_stream_URL', b'pStU'),
    ('current_track', b'pTrk'),
    ('current_visual', b'pVis'),
    ('data', b'pPCT'),
    ('database_ID', b'pDID'),
    ('date_added', b'pAdd'),
    ('description', b'pDes'),
    ('disc_count', b'pDsC'),
    ('disc_number', b'pDsN'),
    ('disliked', b'pHat'),
    ('downloaded', b'pDlA'),
    ('downloader_Apple_ID', b'pDAI'),
    ('downloader_name', b'pDNm'),
    ('duration', b'pDur'),
    ('enabled', b'enbl'),
    ('ending_page', b'lwlp'),
    ('episode_ID', b'pEpD'),
    ('episode_number', b'pEpN'),
    ('EQ', b'pEQp'),
    ('EQ_enabled', b'bpEQ'),
    ('error_handling', b'lweh'),
    ('fax_number', b'faxn'),
    ('finish', b'pStp'),
    ('fixed_indexing', b'pFix'),
    ('format', b'pFmt'),
    ('free_space', b'frsp'),
    ('frontmost', b'pisf'),
    ('full_screen', b'pFSc'),
    ('gapless', b'pGpl'),
    ('genius', b'pGns'),
    ('genre', b'pGen'),
    ('grouping', b'pGrp'),
    ('id', b'bID '),
    ('index', b'pidx'),
    ('kind', b'pKnd'),
    ('location', b'pLoc'),
    ('long_description', b'pLds'),
    ('loved', b'pLov'),
    ('lyrics', b'pLyr'),
    ('media_kind', b'pMdK'),
    ('modifiable', b'pMod'),
    ('modification_date', b'asmo'),
    ('movement', b'pMNm'),
    ('movement_count', b'pMvC'),
    ('movement_number', b'pMvN'),
    ('mute', b'pMut'),
    ('name', b'pnam'),
    ('network_address', b'pMAC'),
    ('pages_across', b'lwla'),
    ('pages_down', b'lwld'),
    ('parent', b'pPlP'),
    ('persistent_ID', b'pPIS'),
    ('played_count', b'pPlC'),
    ('played_date', b'pPlD'),
    ('player_position', b'pPos'),
    ('player_state', b'pPlS'),
    ('position', b'ppos'),
    ('preamp', b'pEQA'),
    ('printer_features', b'lwpf'),
    ('properties', b'pALL'),
    ('protected', b'pPro'),
    ('purchaser_Apple_ID', b'pPAI'),
    ('purchaser_name', b'pPNm'),
    ('rating', b'pRte'),
    ('rating_kind', b'pRtk'),
    ('raw_data', b'pRaw'),
    ('release_date', b'pRlD'),
    ('requested_print_time', b'lwqt'),
    ('resizable', b'prsz'),
    ('sample_rate', b'pSRt'),
    ('season_number', b'pSeN'),
    ('selected', b'selc'),
    ('selection', b'sele'),
    ('shared', b'pShr'),
    ('show', b'pShw'),
    ('shufflable', b'pSfa'),
    ('shuffle', b'pShf'),
    ('shuffle_enabled', b'pShE'),
    ('shuffle_mode', b'pShM'),
    ('size', b'pSiz'),
    ('skipped_count', b'pSkC'),
    ('skipped_date', b'pSkD'),
    ('smart', b'pSmt'),
    ('song_repeat', b'pRpt'),
    ('sort_album', b'pSAl'),
    ('sort_album_artist', b'pSAA'),
    ('sort_artist', b'pSAr'),
    ('sort_composer', b'pSCm'),
    ('sort_name', b'pSNm'),
    ('sort_show', b'pSSN'),
    ('sound_volume', b'pVol'),
    ('special_kind', b'pSpK'),
    ('start', b'pStr'),
    ('starting_page', b'lwfp'),
    ('supports_audio', b'pAud'),
    ('supports_video', b'pVid'),
    ('target_printer', b'trpr'),
    ('time', b'pTim'),
    ('track_count', b'pTrC'),
    ('track_number', b'pTrN'),
    ('unplayed', b'pUnp'),
    ('update_tracks', b'pUTC'),
    ('version', b'vers'),
    ('video_kind', b'pVdK'),
    ('view', b'pPly'),
    ('visible', b'pvis'),
    ('visual_size', b'pVSz'),
    ('visuals_enabled', b'pVsE'),
    ('volume_adjustment', b'pAdj'),
    ('work', b'pWrk'),
    ('year', b'pYr '),
    ('zoomable', b'iszm'),
    ('zoomed', b'pzum'),
]

elements = [
    ('AirPlay_devices', b'cAPD'),
    ('applications', b'capp'),
    ('artworks', b'cArt'),
    ('audio_CD_playlists', b'cCDP'),
    ('audio_CD_tracks', b'cCDT'),
    ('browser_windows', b'cBrW'),
    ('encoders', b'cEnc'),
    ('EQ_presets', b'cEQP'),
    ('EQ_windows', b'cEQW'),
    ('file_tracks', b'cFlT'),
    ('folder_playlists', b'cFoP'),
    ('items', b'cobj'),
    ('library_playlists', b'cLiP'),
    ('miniplayer_windows', b'cMPW'),
    ('playlists', b'cPly'),
    ('playlist_windows', b'cPlW'),
    ('radio_tuner_playlists', b'cRTP'),
    ('shared_tracks', b'cShT'),
    ('sources', b'cSrc'),
    ('subscription_playlists', b'cSuP'),
    ('tracks', b'cTrk'),
    ('URL_tracks', b'cURT'),
    ('user_playlists', b'cUsP'),
    ('video_windows', b'cNPW'),
    ('visuals', b'cVis'),
    ('windows', b'cwin')
]

commands = [
    ('exists', b'coredoex', []),
    ('move', b'coremove', [('to', b'insh')]),
    ('subscribe', b'hookpSub', []),
    ('playpause', b'hookPlPs', []),
    ('download', b'hookDwnl', []),
    ('close', b'coreclos', []),
    ('open', b'aevtodoc', []),
    ('select', b'miscslct', []),
    ('open_location', b'GURLGURL', []),
    ('quit', b'aevtquit', []),
    ('pause', b'hookPaus', []),
    ('make', b'corecrel', [
        ('new', b'kocl'),
        ('at', b'insh'),
        ('with_properties', b'prdt')
    ]),
    ('duplicate', b'coreclon', [('to', b'insh')]),
    ('print_', b'aevtpdoc', [
        ('print_dialog', b'pdlg'),
        ('with_properties', b'prdt'),
        ('kind', b'pKnd'),
        ('theme', b'pThm')
    ]),
    ('add', b'hookAdd', [('to', b'insh')]),
    ('rewind', b'hookRwnd', []),
    ('save', b'coresave', []),
    ('play', b'hookPlay', [('once', b'POne')]),
    ('run', b'aevtoapp', []),
    ('resume', b'hookResu', []),
    ('updatePodcast', b'hookUpd1', []),
    ('next_track', b'hookNext', []),
    ('stop', b'hookStop', []),
    ('search', b'hookSrch', [
        ('for_', b'pTrm'),
        ('only', b'pAre')
    ]),
    ('updateAllPodcasts', b'hookUpdp', []),
    ('update', b'hookUpdt', []),
    ('previous_track', b'hookPrev', []),
    ('fast_forward', b'hookFast', []),
    ('count', b'corecnte', [('each', b'kocl')]),
    ('reveal', b'hookRevl', []),
    ('convert', b'hookConv', []),
    ('eject', b'hookEjct', []),
    ('back_track', b'hookBack', []),
    ('refresh', b'hookRfrs', []),
    ('delete', b'coredelo', [])
]



"""
Sqlite database to cache music player track index / filename mappings
"""

import os

from datetime import datetime
from pytz import timezone
from sqlite3 import OperationalError
from systematic.sqlite import SQLiteDatabase

from pytunes import MusicPlayerError

TABLES_SQL = (
    """
    CREATE TABLE IF NOT EXISTS tracks (
        key INTEGER PRIMARY KEY,
        path STRING,
        mtime INTEGER,
        added DATE
    );
    """,
)


class TrackIndexDB(SQLiteDatabase):
    """
    Track index database

    SQlite index for track paths in MacOS music player
    """
    def __init__(self, client, path):
        super().__init__(path, tables_sql=TABLES_SQL)
        self.client = client

    def __repr__(self):
        return '{}'.format(self.db_path)

    def add_track(self, track):
        """
        Add track to index
        """

        try:
            c = self.cursor
            c.execute("""SELECT key, mtime FROM tracks WHERE key=?""", (track.index,))
            res = c.fetchone()
        except OperationalError as e:
            raise MusicPlayerError(e)

        try:
            mtime = os.stat(track.path).st_mtime
        except OSError:
            mtime = None
        except IOError:
            mtime = None

        if res is not None:
            if mtime:
                if res[1] != mtime:
                    c = self.cursor
                    c.execute("""UPDATE tracks set mtime=? WHERE key=?""", (mtime, res[0],))
                    self.commit()
            else:
                c = self.cursor
                c.execute("""DELETE FROM tracks WHERE key=?""", (res[0],))
                self.commit()

        elif mtime is not None:
            added = datetime.now(timezone('UTC'))
            try:
                c = self.cursor
                c.execute("""INSERT INTO tracks VALUES (?, ?, ?, ?)""", (
                    track.index,
                    track.path,
                    mtime,
                    added
                ))
                self.commit()
            except OperationalError as e:
                raise MusicPlayerError(e)

    def lookup_index(self, path):
        """
        Find index for filename

        Raises MusicPlayerError if path was not in database
        """
        path = os.path.realpath(path)
        try:
            c = self.cursor
            c.execute("""SELECT key FROM tracks WHERE path=?""", (path,))
            res = c.fetchone()
        except OperationalError as e:
            raise MusicPlayerError(e)
        if res:
            return res[0]
        else:
            raise MusicPlayerError('Track not in index database: {}'.format(path))

    def cleanup(self, ids):
        """
        Remove unknown IDs

        Removes tracks with ID not provided in list 'ids'
        """
        try:
            c = self.cursor
            c.execute("""DELETE FROM tracks WHERE key not in (?)""", (','.join(ids),))
            self.commit()
        except OperationalError as e:
            raise MusicPlayerError(e)

    def update(self):
        """
        Update index

        Update tracks track sqlite index
        """
        ids = []
        for _i, track in enumerate(self.client.library):
            if track.path is not None:
                self.add_track(track)
                ids.append('{}'.format(track.index))
        self.cleanup(ids)



# -*- coding: utf-8 -*-
"""
Field and order specs
"""

import appscript

SKIPPED_PLAYLISTS = (
    'iTunes U',
    'Music',
    'Movies',
    'Videos',
    'Home Videos',
    'Films',
    'TV Shows',
    'TV Programmes',
    'Audiobooks',
    'Purchased',
    'Books',
    'PDFs',
    'Genius',
)

TRACK_SYS_FIELDS = (
    'id',
    'index',
    'persistent_ID',
)

TRACK_FIELDS = (
    'name',
    'album',
    'album_artist',
    'album_rating',
    'album_rating_kind',
    'artist',
    'bit_rate',
    'bookmark',
    'bookmarkable',
    'bpm',
    'category',
    'comment',
    'compilation',
    'composer',
    'database_ID',
    'date_added',
    'description',
    'disc_count',
    'disc_number',
    'duration',
    'enabled',
    'episode_ID',
    'episode_number',
    'EQ',
    'finish',
    'gapless',
    'genre',
    'grouping',
    'kind',
    'lyrics',
    'modification_date',
    'played_count',
    'played_date',
    'podcast',
    'rating',
    'rating_kind',
    'release_date',
    'sample_rate',
    'season_number',
    'shufflable',
    'skipped_count',
    'skipped_date',
    'show',
    'sort_album',
    'sort_artist',
    'sort_album_artist',
    'sort_name',
    'sort_composer',
    'sort_show',
    'size',
    'start',
    'time',
    'track_count',
    'track_number',
    'unplayed',
    'video_kind',
    'volume_adjustment',
    'year',
    'location',
)

TRACK_INT_FIELDS = (
    'year',
    'id',
    'index',
    'album_rating',
    'rating',
    'disc_count',
    'disc_number',
    'episode_number',
    'played_count',
    'season_number',
    'sample_rate',
    'size',
    'track_count',
    'track_number',
    'volume_adjustment',
)

TRACK_FLOAT_FIELDS = (
    'bpm',
    'bitrate',
    'bit_rate',
    'duration',
    'bookmark',
    'start',
    'finish',
)

TRACK_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
TRACK_DATE_FIELDS = (
    'date_added',
    'modification_date',
    'played_date',
    'release_date',

)

REPEAT_VALUES = {
    'off': appscript.k.off,
    'one': appscript.k.one,
    'all': appscript.k.all,
}

PLAYER_STATE_NAMES = {
    appscript.k.playing: 'playing',
    appscript.k.paused:  'paused',
    appscript.k.stopped: 'stopped',
    appscript.k.off: False,
    appscript.k.off: False,
}




def parse_exported_playlist_data(filename):
    """
    Example how to load playlist information text file exported
    from iTunes (applies to iOS device backups as well).

    This can be used to figure out which songs there were on a iOS device
    playlist, which the silly thing can't do itself. What you need to do
    is to:
    - connect the iOS device
    - select playlist from the iOS device
    - use File -> Library -> Export Playlist to export playlist info to text file
    - repeat for every playlist on iOS

    To import the playlist directly back to iTunes, just use 'Import Playlist' from
    menu and import the text file ... silly buggers!

    When you have the .txt files, you can use this function to reconstruct the
    playlist contents, to for example export the information to a normal .m3u
    playlist of tracks in your library.
    """
    from csv import reader
    tracks = []
    with open(filename, 'r', encoding='utf-16') as fd:
        rows = reader(fd, delimiter='\t')
        fields = next(rows)
        for row in rows:
            track = {}
            for i, field in enumerate(fields):
                track[field] = row[i]
            tracks.append(track)
    return tracks



"""
Python wrapper for MacOS music players (iTunes / Music) using appscript

Example usage:

from pytunes.client import Client
client = Client()

"""

__version__ = '13.0.2'


class MusicPlayerError(Exception):
    pass



"""
Music player status background daemon
"""

import configobj
import os
import redis

from pytunes import MusicPlayerError
from pytunes.status import musicPlayerStatus

DEFAULT_CONFIGURATION_PATH = '~/Library/Application Support/Pytunes/pytunesd.conf'

# Max songs to keep in redis queue
MAX_REDIS_LIST_ITEMS = 8000


class DaemonConfiguration(dict):
    """
    Configuration for music player daemon
    """

    def __init__(self,
                 path=DEFAULT_CONFIGURATION_PATH,
                 logfile=None, redis_host=None, redis_auth=None, redis_key='pytunes/log'):

        self['logfile'] = logfile
        self['redis_host'] = redis_host
        self['redis_auth'] = redis_auth
        self['redis_key'] = redis_key

        path = os.path.expandvars(os.path.expanduser(path))
        if os.path.isfile(path):
            self.update(configobj.ConfigObj(path).items())


class Daemon(object):
    """
    Background daemon for music player status

    Daemon to monitor music player status changes. By default just logs tracks played.
    """

    def __init__(self, logfile=None, redis_host=None, redis_auth=None, redis_key='pytunesd'):
        self.monitor = musicPlayerStatus()
        self.logfile = logfile
        if redis_host:
            self.redis = redis.client.StrictRedis(host=redis_host, password=redis_auth)
        else:
            self.redis = None
        self.redis_key = redis_key

        if self.redis:
            try:
                self.redis.ping()
            except Exception as e:
                print(e)

    def log_track_change(self, details):
        """Log track changes

        Logs timestamp and path of track played to logfile, if defined.
        """
        if self.logfile is not None:
            path = os.path.expandvars(os.path.expanduser(self.logfile))
            try:
                with open(path, 'a') as f:
                    f.write('{} {}\n'.format(details['started'], details['path']))
                    f.flush()
            except OSError as e:
                raise MusicPlayerError('Error writing to {}: {}'.format(self.log_file, e))
            except IOError as e:
                raise MusicPlayerError('Error writing to {}: {}'.format(self.log_file, e))

        if self.redis is not None:
            self.redis.lpush(self.redis_key, details)
            self.redis.ltrim(self.redis_key, MAX_REDIS_LIST_ITEMS)

    def process_track_change(self, status, details):
        """Process track changes

        Receives current status and track details as arguments. By default logs
        track changes, override to do something else.
        """
        if status == 'playing' and details:
            self.log_track_change(details)

    def run(self):
        """Run daemon

        """
        while True:
            status, details = self.monitor.next() # noqa B305
            try:
                self.process_track_change(status, details)
            except MusicPlayerError as e:
                print(e)




import glob
from setuptools import setup, find_packages
from pytunes import __version__

setup(
    name='pytunes',
    author='Ilkka Tuohela',
    author_email='hile@iki.fi',
    version=__version__,
    keywords='iTunes Music control management playlist export',
    description='iTunes / Music CLI control and python API',
    url='https://github.com/hile/pytunes/',
    license='PSF',
    packages=find_packages(),
    python_requires='>3.6.0',
    entry_points={
        'console_scripts': [
            'pytunes=pytunes.bin.pytunes:main',
            'pytunes-playlists=pytunes.bin.pytunes_playlists:main',
            'pytunes-update=pytunes.bin.pytunes_update:main',
            'pytunesd=pytunes.bin.pytunesd:main',
        ],
    },
    setup_requires=['pytest-runner'],
    tests_require=(
        'pytest',
        'pytest-runner',
        'pytest-datafiles',
    ),
    install_requires=(
        'oodi',
        'systematic>=4.8.0',
        'appscript',
        'configobj',
        'darwinist',
        'redis',
        'zeroconf',
    ),
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: Python Software Foundation License',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3',
        'Topic :: System',
        'Topic :: System :: Systems Administration',
    ],
)



"""
Simple abstraction to MacOS music library from python with appscript
"""

import appscript
import mactypes
import os
import re

from systematic.process import Processes

from oodi.library.tree import Tree

from . import MusicPlayerError
from . import terminology
from .constants import (
    TRACK_FIELDS,
    TRACK_SYS_FIELDS,
    TRACK_INT_FIELDS,
    TRACK_FLOAT_FIELDS,
    TRACK_DATE_FIELDS,
    REPEAT_VALUES,
    PLAYER_STATE_NAMES,
    SKIPPED_PLAYLISTS
)
from .database import TrackIndexDB
from .playlist import Playlist

APPS = {
    'iTunes': {
        'binary': '/Applications/iTunes.app/Contents/MacOS/iTunes',
        'data_directory': os.path.join(os.getenv('HOME'), 'Music/Tunes'),
        'music_directory': os.path.join(os.getenv('HOME'), 'Music/Tunes/iTunes Media/Music'),
        'library_path_filename': 'library_path.txt',
        'index_database': 'iTunes Track Index.sqlite'
    },
    'Music': {
        'binary': '/System/Applications/Music.app/Contents/MacOS/Music',
        'data_directory': os.path.join(os.getenv('HOME'), 'Music/Music/Music Library.musiclibrary'),
        'music_directory': os.path.join(os.getenv('HOME'), 'Music/Music/Media/Music'),
        'library_path_filename': 'library_path.txt',
        'index_database': 'Tracks.sqlite'
    }
}


def detect_app():
    """
    Detect application (iTunes or Music)
    """
    for app, config in APPS.items():
        if os.path.isfile(config['binary']):
            return app, config
    raise MusicPlayerError('Error detecting music player application')


class MusicTree(Tree):
    """
    MacOS music tree, representing one folder on disk configured as music player library path
    """
    def __init__(self, tree_path=None):
        app, config = detect_app()

        if tree_path is None:
            tree_path_file = os.path.join(config['data_directory'], config['library_path_filename'])
            if os.path.isfile(tree_path_file):
                try:
                    tree_path = open(tree_path_file, 'r').read().strip()
                except IOError as e:
                    raise MusicPlayerError('Error opening {}: {}'.format(tree_path, e))
                except OSError as e:
                    raise MusicPlayerError('Error opening {}: {}'.format(tree_path, e))

            else:
                tree_path = config['music_directory']

        Tree.__init__(self, path=tree_path)


class Client(object):
    """Music client application

    Singleton access to music player appscript API.

    Raises MusicPlayerError if music player was not running when initialized.
    """
    __instance__ = None

    def __init__(self):
        self.__app_name__ = None
        self.__binary__ = None
        self.__index_database__ = None
        self.__detect_application__()

        if Client.__instance__ is None:
            Client.__instance__ = Client.Instance(self.__app_name__, self.__binary__)

        self.__dict__['_Client__instance__'] = Client.__instance__
        self.indexdb = TrackIndexDB(self, self.__index_database__)

    def __detect_application__(self):
        """
        Detect which application we have
        """
        app, config = detect_app()
        self.__app_name__ = app
        self.__binary__ = config['binary']
        self.__index_database__ = os.path.join(config['data_directory'], config['index_database'])

    @property
    def application(self):
        """MacOS music player appscript Instance

        Returns instance of appscript client
        """
        return self.__instance__.application

    def __getattr__(self, attr):
        try:
            return getattr(self.__instance__, attr)
        except AttributeError as e:
            raise AttributeError(e)

    class Instance(object):
        """MacOS music player appscript app instance

        Singleton instance of MacOS music player appscript app

        """
        def __init__(self, app_name, binary):
            self.__app_name__ = app_name
            self.__binary__ = binary
            self.application = None

        def __connect__(self):
            if self.is_running:
                self.application = appscript.app(self.__app_name__, terms=terminology)
            else:
                self.application = None
                raise MusicPlayerError('{} is not running'.format(self.__app_name__))

        def __getattr__(self, attr):
            if self.application is None:
                self.__connect__()
            try:
                return getattr(self.application, attr)
            except AttributeError as e:
                raise AttributeError('No such client attribute: {}: {}'.format(attr, e))

        @property
        def path(self):
            if self.application is not None:
                return '{}'.format(self.application).replace("""app(u'""", '').replace("""')""", '')
            else:
                return ''

        @property
        def is_running(self):
            """Check if user has music player open

            Check if music player process exists for this user
            """
            for process in Processes().filter(command=self.__binary__):
                if process.userid == os.geteuid():
                    return True
            return False

    @property
    def status(self):
        """Music player play status

        Returns current music player play status string from PLAYER_STATE_NAMES
        """
        if self.application is None:
            self.__connect__()
        s = self.application.player_state.get()
        try:
            return PLAYER_STATE_NAMES[s]
        except KeyError:
            return 'Unknown ({}'.format(s)

    @property
    def current_track(self):
        """Current track

        Returns currently selected music player track or None
        """
        if self.application is None:
            self.__connect__()
        try:
            return Track(self, self.application.current_track())

        except appscript.reference.CommandError:
            return None

    @property
    def repeat(self):
        """Get repeat

        """
        if self.application is None:
            self.__connect__()
        try:
            value = self.application.current_playlist.song_repeat.get()
            for key in REPEAT_VALUES:
                if value == REPEAT_VALUES[key]:
                    return REPEAT_VALUES[key]
        except appscript.reference.CommandError:
            return None

    @repeat.setter
    def repeat(self, value):
        """Set repeat

        Set repeat mode. Valid values are in pytunes.constants.REPEAT_VALUES
        """
        if self.application is None:
            self.__connect__()
        try:
            self.application.current_playlist.song_repeat.set(to=REPEAT_VALUES[value])
        except KeyError:
            raise MusicPlayerError('Invalid repeat value {}'.format(value))

    @property
    def shuffle(self):
        """Get shuffle mode

        """
        if self.application is None:
            self.__connect__()
        try:
            return self.application.shuffle_enabled.get()
        except appscript.reference.CommandError:
            return None

    @shuffle.setter
    def shuffle(self, value):
        """Set shuffle

        Set shuffle mode
        """
        if self.application is None:
            self.__connect__()
        value = value and True or False
        self.application.shuffle_enabled.set(to=value)

    @property
    def volume(self):
        """Get volume

        """
        if self.application is None:
            self.__connect__()
        return self.application.sound_volume.get()

    @volume.setter
    def volume(self, value):
        """Set volume level

        Must be a integer in range 0-100
        """
        if self.application is None:
            self.__connect__()
        if not isinstance(value, int):
            raise MusicPlayerError('Volume adjustment must be integer value')
        if value < 0 or value > 100:
            raise MusicPlayerError('Volume adjustment must be in range 0-100')
        self.application.sound_volume.set(to=value)

    @property
    def library(self):
        """Music player library

        Return configured music player library
        """
        return Playlist(self)

    @property
    def smart_playlists(self):
        """Return user smart playlists

        """
        playlists = []
        if self.application is None:
            self.__connect__()
        for pl in self.application.user_playlists.get():
            name = pl.name.get()
            if name not in SKIPPED_PLAYLISTS and pl.smart.get():
                playlists.append(Playlist(self, name))
        return playlists

    @property
    def playlists(self):
        """Return user playlists

        Skips smart playlists and playlists in pytunes.constants.SKIPPED_PLAYLISTS
        """
        playlists = []
        if self.application is None:
            self.__connect__()
        for pl in self.application.user_playlists.get():
            name = pl.name.get()
            if name in SKIPPED_PLAYLISTS or pl.smart.get():
                continue
            playlists.append(Playlist(self, name))
        return playlists

    def create_playlist(self, name):
        """Create playlist

        Returns created playlist as Playlist object
        """
        if self.application is None:
            self.__connect__()
        self.application.make(
            new=appscript.k.user_playlist,
            at='Playlists',
            with_properties={appscript.k.name: name},
        )
        return Playlist(self, name)

    def delete_playlist(self, name):
        """Delete playlist

        Note: if there are multiple lists with same name, ALL are removed.
        """
        if self.application is None:
            self.__connect__()
        for pl in self.application.user_playlists.get():
            if pl.name.get() == name:
                pl.delete()

    def get_playlist(self, name):
        """Get playlist by name

        """
        if self.application is None:
            self.__connect__()
        for pl in self.application.user_playlists.get():
            if pl.name.get() == name:
                return Playlist(self, name)
        raise MusicPlayerError('No such playlist: {}'.format(name))

    def get_playlist_track(self, playlist, index):
        """Get playlist track by index

        Playlist must be a reference to internal playlist object: this is used
        from Playlist class.
        """
        return Track(self, playlist.file_tracks[index])

    def previous(self):
        """Previous track

        Jump to previous track
        """
        return self.previous_track()

    def next(self):
        """Next track

        Jump to next track

        """
        return self.next_track()

    def jump(self, index):
        """Play track by index

        """
        if self.application is None:
            self.__connect__()
        try:
            self.application.play(self.library.playlist.file_tracks[index])
        except appscript.reference.CommandError:
            raise MusicPlayerError('Invalid library index: {}'.format(index))
        return self.current_track

    def play(self, path=None):
        """Play track by path

        """
        if self.application is None:
            self.__connect__()
        if path is not None:
            if os.path.isdir(path):
                path = os.path.join(path, os.listdir(path)[0])
            try:
                self.jump(self.indexdb.lookup_index(path))
            except MusicPlayerError:
                pass
            self.application.play(mactypes.Alias(path))
        else:
            self.application.play()
        return self.current_track


class Track(object):
    """
    Track in music player library
    """

    def __init__(self, client, track):
        self.client = client
        self.track = track

        try:
            self.path = self.client.get(self.track.location).path
        except AttributeError:
            self.path = None
        except appscript.reference.CommandError:
            self.path = None

    def __repr__(self):
        return '{} - {} - {}'.format(
            self.artist,
            self.album,
            self.title,
        )

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            pass
        raise AttributeError('Invalid Track attribute: {}'.format(attr))

    def __getitem__(self, item):
        if item == 'path':
            return self.path

        if item == 'extension':
            return os.path.splitext(self.path)[1][1:]

        if item in ('id', 'ID'):
            item = 'id'

        if item in ('date', 'year'):
            item = 'year'

        if item in ('title', 'name'):
            item = 'name'

        try:
            value = self.client.get(self.track.__getattr__(item))
            if value == appscript.k.missing_value:
                value = None

            try:
                if item in TRACK_INT_FIELDS:
                    return int(value)
                elif item in TRACK_FLOAT_FIELDS:
                    return float(value)
                elif item in TRACK_DATE_FIELDS:
                    return value
                elif item == 'location':
                    return value.path
                return str(value)

            except AttributeError:
                return value

        except AttributeError:
            pass

        except appscript.reference.CommandError:
            return None

        raise KeyError('Invalid Track item: {}'.format(item))

    def __setattr__(self, item, value):
        if item in ('date', 'year'):
            item = 'year'
            if type(value) != int:
                try:
                    value = int(value.split('-')[0])
                except ValueError:
                    raise AttributeError('tag {} invalid value: {}'.format(item, value))
                except IndexError:
                    value = int(value)

        if item in TRACK_FIELDS:
            try:
                entry = self.track.__getattr__(item)
                self.client.set(entry, to=value)
            except appscript.reference.CommandError as e:
                raise ValueError('ERROR setting {} to {}: {}'.format(item, value, e))

        elif item in TRACK_SYS_FIELDS:
            raise ValueError('Track attribute {} is read-only'.format(item))

        elif item in ('client', 'track', 'path', 'log'):
            self.__dict__[item] = value

        else:
            raise AttributeError('Invalid Track attribute: {}'.format(item))

    def updateTracknumber(self):
        """Update tracknumber

        Update tracknumber in player. Sets album total tracks to number of files with same extension
        in same directory.
        """
        m = re.match('^([0-9]+) .*', os.path.basename(self.path))
        if not m:
            return False

        track = int(m.group(1))
        name, ext = os.path.splitext(os.path.basename(self.path))
        total = len([f for f in os.listdir(os.path.dirname(self.path)) if os.path.splitext(f)[1] == ext])

        try:
            if self.track_count != total:
                self.track_count = total
        except AttributeError:
            self.track_count = total

        try:
            if self.track_number != track:
                self.track_number = track
        except AttributeError:
            self.track_number = track

        return True

    def updateTag(self, tag, value):
        """Set tag in player metadata

        Set tag to value in player metadata for track
        """
        try:
            current_value = self.__getattr__(tag)
            if tag in ('year', 'date'):
                if type(value) != int:
                    try:
                        value = value.split('-')[0]
                    except IndexError:
                        value = str(value)

            if current_value == value:
                return

        except AttributeError:
            pass

        if value in ('', None,):
            return

        self.__setattr__(tag, value)

    def syncTags(self, song):
        """Sync file tags

        Sync tags from file metadata to player metadata
        """
        modified = False
        try:
            for tag in ('artist', 'album', 'title', 'genre', 'date', 'bpm'):
                if tag not in song.tags:
                    continue

                if tag in ('bpm', 'date'):
                    value = int(song.tags[tag])
                else:
                    value = song.tags[tag]

                if tag == 'bpm' and value == 0:
                    continue

                if tag == 'title':
                    tag = 'name'

                self.updateTag(tag, value)
                modified = True

        except TagError as e:
            raise MusicPlayerError('Error updating tags: {} {}'.format(self.path, e))

        if self.updateTracknumber():
            modified = True

        return modified

    def refresh(self):
        """Refresh trck

        """
        self.client.refresh(self.track)

    def keys(self):
        keys = ['path']
        keys.extend(TRACK_SYS_FIELDS)
        keys.extend(TRACK_FIELDS)
        return keys

    def items(self):
        items = []
        for k in self.keys():
            try:
                items.append([k, self.__getattr__(k)])
            except appscript.reference.CommandError:
                pass
        return items



"""
ITC Albumart files
"""

import os
from struct import unpack
from PIL import ImageFile


class ITCArtworkFile(object):
    """ITC Albumart

    ITC Albumart file parser
    """
    def __init__(self, path):
        self.path = path

        self.library_id, self.track_id = os.path.splitext(
            os.path.basename(path)
        )[0].split('-')

        data = open(self.path, 'r').read()

        hdr_len = unpack('>i', data[0:4])[0]
        self.itc_type = data[hdr_len + 44:hdr_len + 48]
        imagedata = data[hdr_len + 208:]
        parser = ImageFile.Parser()
        parser.feed(imagedata)
        self.image = parser.close()

    def __str__(self):
        return '{0} track {1} {2}'.format(
            self.itc_type,
            self.track_id,
            self.image.format.lower(),
        )

    def write(self, outputdir, filename=None):
        """Write ITC to file

        Write ITC image output file specified directory
        """
        if not filename:
            filename = os.path.join(
                outputdir, '{0}.{1}'.format(
                    self.track_id,
                    self.image.format.lower(),
                )
            )

        if self.image.mode != 'RGB':
            self.image = self.image.convert('RGB')

        self.image.save(filename)
